@@ -20,19 +20,21 @@ import os
 import fnmatch
 import urllib.request
 import ssl
+import yaml
 from apt_inst import DebFile
 from io import BytesIO
 
 import zlib
 import cairo
 from gi.repository import Rsvg
-from tempfile import NamedTemporaryFile
 from PIL import Image
 import logging as log
 
 from dep11.component import DEP11Component, IconSize
 from dep11.parsers import read_desktop_data, read_appstream_upstream_xml
 from dep11.iconfinder import AbstractIconFinder
+from dep11.datacache import DataCache
+
 
 xdg_icon_sizes = [IconSize(64), IconSize(72), IconSize(96), IconSize(128),
                     IconSize(256), IconSize(512)]
@@ -42,14 +44,15 @@ class MetadataExtractor:
     Takes a deb file and extracts component metadata from it.
     '''
 
-    def __init__(self, suite_name, component, export_dir, base_url, icon_sizes, icon_finder=None):
+    def __init__(self, suite_name, component, icon_sizes, dcache, icon_finder=None):
         '''
         Initialize the object with List of files.
         '''
         self._suite_name = suite_name
         self._archive_component = component
-        self._export_dir = export_dir
-        self._base_url = base_url
+        self._export_dir = dcache.media_dir
+        self._dcache = dcache
+        self.write_to_cache = True
 
         self._icon_ext_allowed = ('.png', '.svg', '.xcf', '.gif', '.svgz', '.jpg')
 
@@ -78,6 +81,18 @@ class MetadataExtractor:
     def icon_finder(self, val):
         self._icon_finder = val
 
+    def reopen_cache(self):
+        self._dcache.reopen()
+
+    def get_path_for_cpt(self, cpt, basepath, subdir):
+        gid = cpt.global_id
+        if not gid:
+            return None
+        if len(cpt.cid) < 1:
+            return None
+        path = os.path.join(basepath, gid, subdir)
+        return path
+
     def _get_deb_filelist(self, deb):
         '''
         Returns a list of all files in a deb package
@@ -99,7 +114,7 @@ class MetadataExtractor:
         '''
         thumbnails = list()
         name = os.path.basename(imgsrc)
-        sizes = ['752x423', '624x351', '112x63']
+        sizes = ['1248x702', '752x423', '624x351', '112x63']
         for size in sizes:
             wd, ht = size.split('x')
             img = Image.open(imgsrc)
@@ -114,14 +129,14 @@ class MetadataExtractor:
 
         return thumbnails
 
-    def _fetch_screenshots(self, cpt, cpt_export_path, cpt_public_url):
+    def _fetch_screenshots(self, cpt, cpt_export_path, cpt_public_url=""):
         '''
         Fetches screenshots from the given url and
         stores it in png format.
         '''
 
         if not cpt.screenshots:
-            # don't ignore metadata if screenshots itself is not present
+            # don't ignore metadata if no screenshots are present
             return True
 
         success = True
@@ -133,9 +148,9 @@ class MetadataExtractor:
             if not origin_url:
                 # url empty? skip this screenshot
                 continue
-            path = os.path.join(cpt_export_path, "screenshots")
-            base_url = os.path.join(cpt_public_url, "screenshots")
-            imgsrc = os.path.join(path, "source", "screenshot-%s.png" % (str(cnt)))
+            path     = self.get_path_for_cpt(cpt, cpt_export_path, "screenshots")
+            base_url = self.get_path_for_cpt(cpt, cpt_public_url,  "screenshots")
+            imgsrc   = os.path.join(path, "source", "scr-%s.png" % (str(cnt)))
 
             # The Debian services use a custom setup for SSL verification, not trusting global CAs and
             # only Debian itself. If we are running on such a setup, ensure we load the global CA certs
@@ -174,7 +189,7 @@ class MetadataExtractor:
                 wd, ht = img.size
                 shot['source-image']['width'] = wd
                 shot['source-image']['height'] = ht
-                shot['source-image']['url'] = os.path.join(base_url, "source", "screenshot-%s.png" % (str(cnt)))
+                shot['source-image']['url'] = os.path.join(base_url, "source", "scr-%s.png" % (str(cnt)))
                 img.close()
             except Exception as e:
                 error_msg = str(e)
@@ -218,9 +233,11 @@ class MetadataExtractor:
 
         img.write_to_png(store_path)
 
-    def _store_icon(self, cpt, cpt_export_path, icon_path, deb_fname, size):
+    def _store_icon(self, deb_fname, cpt, cpt_export_path, icon_path, size):
         '''
         Extracts the icon from the deb package and stores it in the cache.
+        Ensures the stored icon always has the size given in "size", and renders
+        vectorgraphics if necessary.
         '''
         svgicon = False
         if not self._icon_allowed(icon_path):
@@ -230,7 +247,7 @@ class MetadataExtractor:
         if not os.path.exists(deb_fname):
             return False
 
-        path = "%s/icons/%s/" % (cpt_export_path, str(size))
+        path = self.get_path_for_cpt(cpt, cpt_export_path, "icons/%s" % (str(size)))
         icon_name = "%s_%s" % (cpt.pkgname, os.path.basename(icon_path))
         icon_name_orig = icon_name
 
@@ -246,12 +263,18 @@ class MetadataExtractor:
 
         # filepath is checked because icon can reside in another binary
         # eg amarok's icon is in amarok-data
+        icon_data = None
         try:
             icon_data = DebFile(deb_fname).data.extractdata(icon_path)
         except Exception as e:
             cpt.add_hint("deb-extract-error", {'fname': icon_name, 'pkg_fname': deb_fname, 'error': str(e)})
             return False
 
+        if not icon_data:
+            cpt.add_hint("deb-extract-error", {'fname': icon_name, 'pkg_fname': deb_fname,
+                                               'error': "Icon data was empty. The icon might be a symbolic link, please do not symlink icons "
+                                                         "(instead place the icons in their appropriate directories in <code>/usr/share/icons/hicolor/</code>)."})
+            return False
         cpt.icon = icon_name
 
         if icon_name_orig.endswith(".svg"):
@@ -264,49 +287,59 @@ class MetadataExtractor:
                 cpt.add_hint("svgz-decompress-error", {'icon_fname': icon_name, 'error': str(e)})
                 return False
 
-        if icon_data:
-            if not os.path.exists(path):
-                os.makedirs(os.path.dirname(path))
+        if not os.path.exists(path):
+            os.makedirs(path)
 
-            if svgicon:
-                # render the SVG to a bitmap
-                self._render_svg_to_png(icon_data, icon_store_location, int(size), int(size))
-                return True
-            else:
-                # we don't trust upstream to have the right icon size present, and therefore
-                # always adjust the icon to the right size
-                stream = BytesIO(icon_data)
-                stream.seek(0)
-                img = None
-                try:
-                    img = Image.open(stream)
-                except Exception as e:
-                    cpt.add_hint("icon-open-failed", {'icon_fname': icon_name, 'error': str(e)})
-                    return False
-                newimg = img.resize((int(size), int(size)), Image.ANTIALIAS)
-                newimg.save(icon_store_location)
-                return True
+        if svgicon:
+            # render the SVG to a bitmap
+            self._render_svg_to_png(icon_data, icon_store_location, int(size), int(size))
+            return True
+        else:
+            # we don't trust upstream to have the right icon size present, and therefore
+            # always adjust the icon to the right size
+            stream = BytesIO(icon_data)
+            stream.seek(0)
+            img = None
+            try:
+                img = Image.open(stream)
+            except Exception as e:
+                cpt.add_hint("icon-open-failed", {'icon_fname': icon_name, 'error': str(e)})
+                return False
+            newimg = img.resize((int(size), int(size)), Image.ANTIALIAS)
+            newimg.save(icon_store_location)
+            return True
 
         return False
 
-    def _match_and_store_icon(self, cpt, cpt_export_path, pkg_fname, filelist, icon_name, size):
-        success = False
+
+    def _match_icon_on_filelist(self, cpt, filelist, icon_name, size):
         if size == "scalable":
             size_str = "scalable"
         else:
             size_str = str(size)
-        icon_path = "usr/share/icons/hicolor/%s/*/%s" % (size_str, icon_name)
+        icon_path = "usr/share/icons/hicolor/%s/apps/%s" % (size_str, icon_name)
         filtered = fnmatch.filter(filelist, icon_path)
         if not filtered:
+            return None
+
+        return filtered[0]
+
+
+    def _match_and_store_icon(self, pkg_fname, cpt, cpt_export_path, filelist, icon_name, size):
+        success = False
+        matched_icon = self._match_icon_on_filelist(cpt, filelist, icon_name, size)
+        if not matched_icon:
             return False
+
         if not size in self._icon_sizes:
             # scale icons to allowed sizes
             for asize in self._icon_sizes:
-                success = self._store_icon(cpt, cpt_export_path, filtered[0], pkg_fname, asize) or success
+                success = self._store_icon(pkg_fname, cpt, cpt_export_path, matched_icon, asize) or success
         else:
-            success = self._store_icon(cpt, cpt_export_path, filtered[0], pkg_fname, size)
+            success = self._store_icon(pkg_fname, cpt, cpt_export_path, matched_icon, size)
         return success
 
+
     def _fetch_icon(self, cpt, cpt_export_path, pkg_fname, filelist):
         '''
         Searches for icon if absolute path to an icon
@@ -319,10 +352,13 @@ class MetadataExtractor:
         icon_str = cpt.icon
         cpt.icon = None
 
+        all_icon_sizes = self._icon_sizes[:]
+        all_icon_sizes.extend(self._large_icon_sizes)
+
         success = False
         if icon_str.startswith("/"):
             if icon_str[1:] in filelist:
-                return self._store_icon(cpt, cpt_export_path, icon_str[1:], pkg_fname, IconSize(64))
+                return self._store_icon(pkg_fname, cpt, cpt_export_path, icon_str[1:], IconSize(64))
         else:
             ret = False
             icon_str = os.path.basename (icon_str)
@@ -332,18 +368,46 @@ class MetadataExtractor:
                 icon_name_ext = icon_str
             else:
                 icon_name_ext = icon_str + ".png"
+
+            found_sizes = list()
             for size in self._icon_sizes:
-                success = self._match_and_store_icon(cpt, cpt_export_path, pkg_fname, filelist, icon_name_ext, size) or success
+                ret = self._match_and_store_icon(pkg_fname, cpt, cpt_export_path, filelist, icon_name_ext, size)
+                if ret:
+                    found_sizes.append(size)
+                success = ret or success
+
+            # try if we can add missing icon sizes by scaling down things
+            # this also ensures that we also have an 64x64 sized icon
+            if set(found_sizes) != set(self._icon_sizes):
+                for size in self._icon_sizes:
+                    if size in found_sizes:
+                        continue
+                    for asize in all_icon_sizes:
+                        if asize < size:
+                            continue
+                        icon_fname = self._match_icon_on_filelist(cpt, filelist, icon_name_ext, asize)
+                        if not icon_fname:
+                            continue
+                        ret = self._store_icon(pkg_fname, cpt, cpt_export_path, icon_fname, size)
+                        if ret:
+                            found_sizes.append(size)
+                        success = ret or success
+                        break
+
+            # a 64x64 icon is required, so double-check if we have one
+            if success and not IconSize(64) in found_sizes:
+                success = False
+
             if not success:
                 # we cheat and test for larger icons as well, which can be scaled down
                 # first check for a scalable graphic
-                success = self._match_and_store_icon(cpt, cpt_export_path, pkg_fname, filelist, icon_str + ".svg", "scalable")
+                success = self._match_and_store_icon(pkg_fname, cpt, cpt_export_path, filelist, icon_str + ".svg", "scalable")
                 if not success:
-                    success = self._match_and_store_icon(cpt, cpt_export_path, pkg_fname, filelist, icon_str + ".svgz", "scalable")
+                    success = self._match_and_store_icon(pkg_fname, cpt, cpt_export_path, filelist, icon_str + ".svgz", "scalable")
                 # then try to scale down larger graphics
                 if not success:
                     for size in self._large_icon_sizes:
-                        success = self._match_and_store_icon(cpt, cpt_export_path, pkg_fname, filelist, icon_name_ext, size) or success
+                        success = self._match_and_store_icon(pkg_fname, cpt, cpt_export_path, filelist, icon_name_ext, size) or success
 
         if not success:
             last_pixmap = None
@@ -355,7 +419,7 @@ class MetadataExtractor:
                         # the pixmap dir can contain icons in multiple formats, and store_icon() fails in case
                         # the icon format is not allowed. We therefore only exit here, if the icon has a valid format
                         if self._icon_allowed(path):
-                            return self._store_icon(cpt, cpt_export_path, path, pkg_fname, IconSize(64))
+                            return self._store_icon(pkg_fname, cpt, cpt_export_path, path, IconSize(64))
                         last_pixmap = path
             if last_pixmap:
                 # we don't do a global icon search anymore, since we've found an (unsuitable) icon
@@ -363,8 +427,6 @@ class MetadataExtractor:
                 cpt.add_hint("icon-format-unsupported", {'icon_fname': os.path.basename(last_pixmap)})
                 return False
 
-            all_icon_sizes = self._icon_sizes[:]
-            all_icon_sizes.extend(self._large_icon_sizes)
             icon_dict = self._icon_finder.find_icons(cpt.pkgname, icon_str, all_icon_sizes)
             success = False
             if icon_dict:
@@ -372,20 +434,20 @@ class MetadataExtractor:
                     if not size in icon_dict:
                         continue
 
-                    success = self._store_icon(cpt,
+                    success = self._store_icon(icon_dict[size]['deb_fname'],
+                                        cpt,
                                         cpt_export_path,
                                         icon_dict[size]['icon_fname'],
-                                        icon_dict[size]['deb_fname'],
                                         size) or success
                 if not success:
                     for size in self._large_icon_sizes:
                         if not size in icon_dict:
                             continue
                         for asize in self._icon_sizes:
-                            success = self._store_icon(cpt,
+                            success = self._store_icon(icon_dict[size]['deb_fname'],
+                                        cpt,
                                         cpt_export_path,
                                         icon_dict[size]['icon_fname'],
-                                        icon_dict[size]['deb_fname'],
                                         asize) or success
                 return success
 
@@ -395,7 +457,8 @@ class MetadataExtractor:
                 cpt.add_hint("icon-not-found", {'icon_fname': icon_str})
             return False
 
-        return True
+        return success
+
 
     def process(self, pkgname, pkg_fname, pkgid=None, metainfo_files=None):
         '''
@@ -421,18 +484,14 @@ class MetadataExtractor:
             cpt.add_hint("deb-filelist-error", {'pkg_fname': os.path.basename(pkg_fname)})
             return [cpt]
 
-        component_basepath = None
-        if pkgid:
-            component_basepath = "%s/%s" % (self._archive_component, pkgid)
-        else:
+        if not pkgid:
             # we didn't get an identifier, so start guessing one.
             idname, ext = os.path.splitext(os.path.basename(pkg_fname))
             if not idname:
                 idname = os.path.basename(pkg_fname)
-            component_basepath = "%s/%s" % (self._archive_component, idname)
-        export_path = "%s/%s" % (self._export_dir, component_basepath)
-        public_url = "%s/%s" % (self._base_url, component_basepath)
+            pkgid = idname
 
+        export_path = "%s/%s" % (self._export_dir, self._archive_component)
         component_dict = dict()
 
         # if we don't have an explicit list of interesting files, we simply scan all
@@ -449,7 +508,7 @@ class MetadataExtractor:
 
                 error = None
                 try:
-                    dcontent = str(deb.data.extractdata(meta_file).decode("utf-8"))
+                    dcontent = str(deb.data.extractdata(meta_file), 'utf-8')
                 except Exception as e:
                     error = {'tag': "deb-extract-error",
                                 'params': {'fname': cpt_id, 'pkg_fname': os.path.basename(pkg_fname), 'error': str(e)}}
@@ -465,7 +524,7 @@ class MetadataExtractor:
                 cpt = DEP11Component(self._suite_name, self._archive_component, pkgname, pkgid)
 
                 try:
-                    xml_content = str(deb.data.extractdata(meta_file).decode("utf-8"))
+                    xml_content = str(deb.data.extractdata(meta_file), 'utf-8')
                 except Exception as e:
                     # inability to read an AppStream XML file is a valid reason to skip the whole package
                     cpt.add_hint("deb-extract-error", {'fname': meta_file, 'pkg_fname': os.path.basename(pkg_fname), 'error': str(e)})
@@ -474,13 +533,15 @@ class MetadataExtractor:
                     continue
 
                 read_appstream_upstream_xml(cpt, xml_content)
+                component_dict[cpt.cid] = cpt
+
                 # Reads the desktop files associated with the xml file
                 if not cpt.cid:
                     # if there is no ID at all, we dump this component, since we cannot do anything with it at all
                     cpt.add_hint("metainfo-no-id")
                     continue
 
-                component_dict[cpt.cid] = cpt
+                cpt.set_srcdata_checksum_from_data(xml_content)
                 if cpt.kind == "desktop-app":
                     data = mdata_raw.get(cpt.cid)
                     if not data:
@@ -492,6 +553,7 @@ class MetadataExtractor:
                     else:
                         # we have a .desktop component, extend it with the associated .desktop data
                         read_desktop_data(cpt, data['data'])
+                        cpt.set_srcdata_checksum_from_data(xml_content+data['data'])
                     del mdata_raw[cpt.cid]
 
         # now process the remaining metadata files, which have not been processed together with the XML
@@ -509,16 +571,50 @@ class MetadataExtractor:
                     ret = read_desktop_data(cpt, mdata['data'])
                     if ret or not cpt.has_ignore_reason():
                         component_dict[cpt.cid] = cpt
+                        cpt.set_srcdata_checksum_from_data(mdata['data'])
                     else:
                         # this means that reading the .desktop file failed and we should
                         # silently ignore this issue (since the file was marked to be invisible on purpose)
                         pass
 
-        for cpt in component_dict.values():
+        # fetch media (icons/screenshots), if we don't ignore the component already
+        cpts = component_dict.values()
+        for cpt in cpts:
+            if cpt.has_ignore_reason():
+                continue
+            if not cpt.global_id:
+                log.error("Component '%s' from package '%s' has no source-data checksum / global-id." % (cpt.cid, pkg_fname))
+                continue
+
+            # check if we have a component generated from
+            # this source data in the cache already.
+            # To account for packages which change their package name, we
+            # also need to check if the package this component is associated
+            # with matches ours.
+            existing_mdata = self._dcache.get_metadata(cpt.global_id)
+            if existing_mdata:
+                s = "Package: %s\n" % (pkgname)
+                if s in existing_mdata:
+                    continue
+                else:
+                    # the exact same metadata exists in a different package already, raise ab error.
+                    # ATTENTION: This does not cover the case where *different* metadata (as in, different summary etc.)
+                    # but with the *same ID* exists. This kind of issue can only be catched when listing all IDs per
+                    # suite/acomponent combination and checking for dupes (we do that in the DEP-11 validator and display
+                    # the result prominently on the HTML pages)
+                    ecpt = yaml.safe_load(existing_mdata)
+                    cpt.add_hint("metainfo-duplicate-id", {'cid': cpt.cid, 'pkgname': ecpt.get('Package', '')})
+                    continue
+
             self._fetch_icon(cpt, export_path, pkg_fname, filelist)
             if cpt.kind == 'desktop-app' and not cpt.icon:
                 cpt.add_hint("gui-app-without-icon", {'cid': cpt.cid})
             else:
-                self._fetch_screenshots(cpt, export_path, public_url)
+                self._fetch_screenshots(cpt, export_path)
 
-        return component_dict.values()
+        # write data to cache
+        if self.write_to_cache:
+            # write the components we found to the cache
+            self._dcache.set_components(pkgid, cpts)
+
+        return cpts
